from pathlib import Path
from typing import List
from Bio.Align import MultipleSeqAlignment

from lxml import etree as ET
from xml.etree.ElementTree import ElementTree
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from datetime import datetime


class BeastXML:
    """Class for editing BEAST XML files."""

    file_name: Path
    traits: List[dict]
    xml: ElementTree
    date_trait: bool
    date_format: str
    date_delimiter: str

    def __init__(
        self,
        file_name: Path,
        traits: List[dict] = [],
        date_trait: bool = True,
        date_format: str = "%Y-%m-%d",
        date_delimiter: str = "_",
    ):
        self.file_name = file_name
        self.xml = self._load_xml()
        self.traits = traits
        for trait in traits:
            try:
                self._get_first_element_by_attribute("traitname", trait)
            except ValueError:
                raise ValueError(f"Could not find '{trait}' trait in xml.")
        # should check to see if there are traits in the xml
        # that are not in self.traits?
        if date_trait:
            try:
                self._get_first_element_by_attribute("traitname", "date")
            except ValueError:
                raise ValueError(f"Could not find 'date' trait in xml..")
        self.date_trait = date_trait
        self.date_format = date_format
        self.date_delimiter = date_delimiter

    def _load_xml(self):
        return ET.parse(str(self.file_name))

    def _get_first_element_by_attribute(self, attr, value):
        el = self.xml.find(f".//*[@{attr}='{value}']")
        if el is None:
            raise ValueError(f"Could not find trait with {attr}='{value}'")
        return el

    def _add_trait(self, sequence_id, traitname):
        trait_el = self._get_first_element_by_attribute("traitname", traitname)
        trait = self.get_trait_data(sequence_id, traitname)
        trait_el.set("value", f"{trait_el.get('value')},{sequence_id}={trait}")

    @property
    def alignment(self) -> MultipleSeqAlignment:
        msa = MultipleSeqAlignment([])
        data = self.xml.find("data")
        for sequence_el in data:
            msa.append(
                SeqRecord(
                    Seq(sequence_el.get("value")),
                    id=sequence_el.get("taxon"),
                    description="",
                )
            )
        msa.sort()
        return msa

    def get_sequence_ids(self) -> list:
        return [s.id for s in self.alignment]

    def get_trait_data(self, sequence_id, traitname):
        if traitname == "date" and self.date_trait:
            date = None
            for potential_date in sequence_id.split(self.date_delimiter):
                try:
                    date = datetime.strptime(potential_date, self.date_format).strftime(
                        self.date_format
                    )
                except ValueError:
                    pass
            if not date:
                raise ValueError(
                    f"Could not parse date trait with date format '{self.date_format}' and date delimiter '{self.date_delimiter}'"
                )
            return date

        trait = next(t for t in self.traits if t["traitname"] == traitname)

        return sequence_id.split(trait["delimiter"])[trait["group"]]

    def add_sequence(self, record: SeqRecord):
        if record.id in self.get_sequence_ids():
            raise ValueError(f"New sequence id must be unique ({record.id})")
        try:
            self.alignment.append(record)
        except ValueError as e:
            raise e
        if self.date_trait:
            self._add_trait(record.id, "date")
        for trait in self.traits:
            self._add_trait(record.id, trait["traitname"])
        data = self.xml.find("data")
        sequence_el = ET.Element(
            "sequence",
            {
                "id": f"seq_{record.id}",
                "spec": "Sequence",
                "taxon": record.id,
                "totalcount": "4",
                "value": str(record.seq),
            },
        )
        sequence_el.tail = "\n"
        data.append(sequence_el)
        data[:] = sorted(data, key=lambda x: x.get("taxon"))

    def write(self, out_file=None) -> None:
        if not out_file:
            out_file = self.file_name
        ET.indent(self.xml, space="    ", level=0)
        with open(out_file, "w") as f:
            xml_string = ET.tostring(self.xml, pretty_print=True, encoding=str)
            f.write(xml_string)



from pathlib import Path
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.Align import MultipleSeqAlignment
import typer
from typing import List, Optional, Tuple
from datetime import datetime

from .xml import BeastXML
from .state import StateTree

app = typer.Typer()


def find_closest_sequence(
    sequences_from_xml: MultipleSeqAlignment, new_sequence_record: SeqRecord
):
    max_score = None
    seq_id = None
    with typer.progressbar(sequences_from_xml) as progress:
        for i, sequence in enumerate(progress):

            if sequence.id == new_sequence_record.id:
                continue
            score = sum(
                xi != yi
                for xi, yi in zip(str(sequence.seq), str(new_sequence_record.seq))
            )
            if max_score == None:
                max_score = score
                seq_id = i
            elif score < max_score:
                max_score = score
                seq_id = i
    if seq_id == None:
        raise Exception("No Seq found?")
    return seq_id, max_score


def get_sequences_to_add(fasta_file, list_of_seq_ids: list):
    records = SeqIO.parse(fasta_file, "fasta")
    return [record for record in records if record.id not in list_of_seq_ids]


def decimal_year(dt: datetime):
    year_part = dt - datetime(year=dt.year, month=1, day=1)
    year_length = datetime(year=dt.year + 1, month=1, day=1) - datetime(
        year=dt.year, month=1, day=1
    )
    return dt.year + year_part / year_length


@app.command()
def main(
    xml_file: Path,
    fasta_file: Path,
    state_file: Path = None,
    output: Path = None,
    template: bool = typer.Option(
        False,
        help="XML file is a beast template. New sequences must be append to the end of the original fasta file.",
    ),
    date_trait: bool = True,
    date_format: str = "%Y-%m-%d",
    date_delimiter: str = "_",
    trait: Optional[List[str]] = typer.Option(
        None,
        help="Trait information 'traitname delimiter group' string seperated by spaces",
    ),
):
    if not state_file:
        state_file = Path(f"{xml_file}.state")
    state_tree = StateTree(state_file)
    traits = [
        {
            "traitname": t.split(" ")[0],
            "delimiter": t.split(" ")[1],
            "group": int(t.split(" ")[2]),
        }
        for t in trait
    ]

    beast_xml = BeastXML(
        xml_file,
        traits,
        date_trait=date_trait,
        date_format=date_format,
        date_delimiter=date_delimiter,
    )
    if template:
        date_trait = False
        records = list(SeqIO.parse(fasta_file, "fasta"))
        number_of_tips = state_tree.tree.count_terminals()
        alignment = MultipleSeqAlignment(records[:-number_of_tips])
        sequences_to_add = records[number_of_tips:]
    else:
        sequences_to_add = get_sequences_to_add(
            fasta_file, beast_xml.get_sequence_ids()
        )

    if not sequences_to_add:
        typer.echo("No new sequences found in the fasta file.")
        raise typer.Exit(code=1)

    typer.echo(f"Adding {len(sequences_to_add)} new sequences")
    for sequence in sequences_to_add:
        typer.echo(f"Adding new sequence: {sequence.id}")
        if not template:
            alignment = beast_xml.alignment
        if len(sequence) != alignment.get_alignment_length():
            raise ValueError("Sequences must all be the same length")

        closest_tree_node_id, max_score = find_closest_sequence(alignment, sequence)
        sampling_time_delta = 0
        if date_trait:
            closest_tree_node_date = beast_xml.get_trait_data(
                alignment[closest_tree_node_id].id, traitname="date"
            )
            closest_tree_node_dt = datetime.strptime(
                closest_tree_node_date, beast_xml.date_format
            )
            new_tree_node_date = beast_xml.get_trait_data(sequence.id, traitname="date")
            new_tree_node_dt = datetime.strptime(
                new_tree_node_date, beast_xml.date_format
            )
            sampling_time_delta = decimal_year(new_tree_node_dt) - decimal_year(
                closest_tree_node_dt
            )

        ids = [s.id for s in alignment]
        ids.append(sequence.id)
        index = sorted(ids).index(sequence.id)

        new_clade = state_tree.graft(
            closest_tree_node_id, index=index, sampling_time_delta=sampling_time_delta
        )
        # name = new_clade.name
        # new_clade.name = f"{sequence.id}"
        # state_tree.draw()
        # new_clade.name = name
        if not template:
            beast_xml.add_sequence(sequence)
        else:
            alignment.append(sequence)

    if not template:
        beast_xml.write(out_file=output)
    # else:
    #     for c in state_tree.tree.get_terminals():
    #         c.name = str(int(c.name) + 1)
    state_tree.draw()
    state_tree.write(out_file=output)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['online_beast']

package_data = \
{'': ['*']}

install_requires = \
['biopython>=1.79,<2.0', 'lxml>=4.7.1,<5.0.0', 'typer[all]>0.3.2']

entry_points = \
{'console_scripts': ['online-beast = online_beast.main:app']}

setup_kwargs = {
    'name': 'online-beast',
    'version': '0.7.6',
    'description': '',
    'long_description': '# online-BEAST\n[![PyPi](https://img.shields.io/pypi/v/online-beast.svg)](https://pypi.org/project/online-beast/)\n[![tests](https://github.com/Wytamma/online-beast/actions/workflows/test.yml/badge.svg)](https://github.com/Wytamma/online-beast/actions/workflows/test.yml)\n[![cov](https://codecov.io/gh/Wytamma/online-beast/branch/master/graph/badge.svg)](https://codecov.io/gh/Wytamma/online-beast)\n\nThis command line tool can be used to add sequences to an ongoing analysis in BEAST2. This framework is called online Bayesian phylodynamic inference (see [Gill et al., 2020](https://academic.oup.com/mbe/article/37/6/1832/5758268?login=false)).\n\n## Install\nInstall `online-beast` with pip (requires python -V >= 3.6.2).\n\n```bash\npip install online-beast\n```\n\n## Usage \n\nGive `online-beast` beast the path to a XML file from a previous BEAST2 run (i.e. one that has an associated `.state` file) and a fasta file of sequences to add to the analysis. Sequences in the fasta file must be aligned (i.e. to the sequences in the XML file) and the same length as the other sequences in the XML file. Only new sequences (new descriptors) will be added to the analysis, so new sequences can be append to the fasta file as they are acquired. \n\n```bash\nonline-beast data/testGTR.xml data/samples.fasta\n```\n\n![](images/output.png)\n\nThe new sequences will by added to the XML file and the associated `.state` file (produced automatically by BEAST2).\n\nThe analysis can then be resumed (with the additional sequence data) using the BEAST2 resume flag. \n\n```bash\nbeast -resume testGTR.xml\n```\n\nThe online analysis can be visualised in real-time using [Beastiary](https://beastiary.wytamma.com/). The jumps in the trace show where new sequences have been added. \n\n![](images/beastiary.png)\n\nDate trait data will be automatically parsed. The format of the date trait data (in the fasta descriptor) can be set with the `--date-format` (default `%Y-%m-%d`) and `--delimiter` (default `_`) flags. If there is no date trait in the xml use the `--no-date-trait` flag.\n\n```\nonline-beast data/ebola.xml data/ebola.fasta --dateformat %d/%m/%Y --date-delimiter _\n```\n\nIf there is trait data in the XML file you need to specify how to extract it from the fasta descriptor line using the `--trait` flag. The format is `\'traitname delimiter group\'` e.g. a string separated by spaces. For example to get the `location` trait from `sample_wuhan_2022-04-05` you would use `--trait \'location _ 1\'`. The `--trait` flag can be used multiple times to specify multiple traits. \n\n```bash\nonline-beast covid.xml data/covid.fasta --trait \'location _ 1\'\n```\n\nBy default the new sequences will be appended to the input XML and state files. Output file names can be specified using the `--output` flag. This will also create a new `.state` file.\n\n```bash\nonline-beast testGTR.xml samples.fasta --output new_testGTR.xml \n```\n\nIf you use the BEAST2 `-statefile` flag to specify the filename of the state (i.e. it is not `xml_filename + .state`). Use the flag `--state-file` to specify the state file path. \n\n```bash\nonline-beast testGTR.xml samples.fasta --state-file beast.state \n```\n\n## Explanation\n\nA Markov chain started anywhere near the center of the stationary distribution needs no burn-in ([Geyer 2011](http://www.mcmchandbook.net/)). Online Bayesian phylodynamic inference is akin to transfer learning in the deep learning field. By starting our MCMC with reasonable states (obtained from a previous run) we reduce the amount of optimisation (burn-in) that must be performed to reach convergence. \n\nOnline-beast loosely follows the implementation of [Gill et al., 2020](https://academic.oup.com/mbe/article/37/6/1832/5758268?login=false) for BEAST1. However, most of the implementation of online-beast is handled by the default state system in BEAST2. New sequences are added from the fasta file one at a time. The hamming distance is calculated between the new sequence and all the other sequences in the XML file. The new sequence is grafted onto the tree in the `.state` file, half way along the branch of the closest sequence in the XML file. The new sequence is append to the BEAST XML file. \n\n## Ebola example\n\nIn this example we will make use of a publicly available dataset of sequences from the 2013-2016 *Zaire ebolavirus* outbreak in Sierra Leone. \n\nIn the `data/` folder you\'ll find a `ebola.xml` file and several fasta files that contain sequences from the outbreak broken up by date. The script below will run an online Bayesian phylodynamic analysis adding new sequences after each run finishes. \n\n\n```bash\n#!/bin/bash\n\n# Run beast with initial samples\nbeast data/ebola.xml \n# Update analysis with new samples\nonline-beast data/ebola.xml data/ebola1.fasta --date-format "%d/%m/%Y" --state-file ebola.xml.state --output ebola.xml\n# Resume the analysis\nbeast -resume ebola.xml \n# Update analysis with new samples\nonline-beast ebola.xml data/ebola2.fasta --date-format "%d/%m/%Y" --output ebola.xml\n# Resume the analysis\nbeast -resume ebola.xml \n# Update analysis with new samples\nonline-beast ebola.xml data/ebola3.fasta --date-format "%d/%m/%Y" --output ebola.xml\n# Resume the analysis\nbeast -resume ebola.xml \n```\n\n\n\n\n',
    'author': 'Wytamma Wirth',
    'author_email': 'wytamma.wirth@me.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0',
}


setup(**setup_kwargs)



from pathlib import Path
from Bio import Phylo
from io import StringIO
from Bio.Phylo.BaseTree import TreeMixin


class StateTree:
    """Class for editing BEAST state files."""

    file_name: str
    tree: TreeMixin

    def __init__(self, file_name: Path):
        self.file_name = file_name
        self.tree = self._get_tree()

    def _get_tree(self):
        with open(self.file_name) as f:
            lines = f.readlines()
        tree_line = lines[1]
        newick_tree = tree_line.split(">")[1].split("</")[0]
        return Phylo.read(StringIO(newick_tree), "newick")

    @property
    def newick(self):
        writer = Phylo.NewickIO.Writer([self.tree])
        return next(writer.to_strings(format_branch_length="%1.17f"))

    def get_parent(self, child_clade):
        node_path = self.tree.get_path(child_clade)
        return node_path[-2]

    def graft(
        self,
        branch_id,
        index: int,
        sampling_time_delta: float = 0,
        graft_point: float = 0.5,
    ):
        try:
            clade = next(
                c for c in self.tree.get_terminals() if c.name == str(branch_id)
            )
        except StopIteration:
            raise ValueError("Could not find branch with id", branch_id)
        total_branch_length = clade.branch_length

        internal = False

        starting_total_branch_length = self.tree.total_branch_length()

        while total_branch_length + sampling_time_delta < 0:
            try:
                clade = self.get_parent(clade)
            except IndexError:
                # root
                ValueError("Sample date before root...")
            internal = True
            total_branch_length += clade.branch_length

        if sampling_time_delta < 0:
            parent_branch_length = (
                total_branch_length - abs(sampling_time_delta)
            ) * graft_point
            # parent_branch_length = (clade.branch_length) * graft_point
            og_parent = clade.branch_length
            clade.branch_length = parent_branch_length
            branch_length = total_branch_length - parent_branch_length
        else:
            parent_branch_length = (total_branch_length) * graft_point
            clade.branch_length = parent_branch_length
            branch_length = total_branch_length - parent_branch_length

        for leaf in self.tree.get_terminals():
            if int(leaf.name) < index:
                continue
            leaf.name = str(int(leaf.name) + 1)

        if internal:
            descendants = 1
        else:
            descendants = 2

        clade.split(n=descendants, branch_length=branch_length)
        clade.name = None

        if internal:
            children = list(clade.clades)
            clade.clades = []
            clade.split(branch_length=0)
            clade.clades[1] = children[-1]
            clade.clades[0].branch_length = og_parent - clade.branch_length
            clade.clades[0].clades = children[:-1]
            clade.clades[0].confidence = None
            clade.clades[0].name = None
        else:
            original_branch = clade.clades[0]
            if branch_id >= index:
                branch_id += 1
            original_branch.name = str(branch_id)

        new_branch = clade.clades[1]
        if sampling_time_delta:
            if new_branch.branch_length + sampling_time_delta < 0:
                raise ValueError("negative branch bad :(")
            new_branch.branch_length = new_branch.branch_length + sampling_time_delta

        error = 0.00001  # for floats
        if (
            abs(
                (self.tree.total_branch_length() - new_branch.branch_length)
                - starting_total_branch_length
            )
            > error
        ):
            print(
                self.tree.total_branch_length(),
                new_branch.branch_length,
                (self.tree.total_branch_length() - new_branch.branch_length),
                starting_total_branch_length,
            )
            raise ValueError("Tree length is wrong!")

        if index != None:
            new_branch.name = str(index)
        else:
            new_branch.name = str(len(self.tree.get_terminals()) - 1)  # zero indexed
        return new_branch

    def draw(self):
        Phylo.draw_ascii(self.tree)

    def write(self, out_file=None) -> None:
        with open(self.file_name) as f:
            state_file_lines = f.readlines()

        for c in self.tree.find_clades():
            if c.confidence:
                c.confidence = None

        old_tree_line = state_file_lines[1]
        opening_tag = old_tree_line.split(">")[0]
        closting_tag = old_tree_line.split("</")[-1]
        state_file_lines[1] = f"{opening_tag}>{self.newick}</{closting_tag}"

        if not out_file:
            out_file = self.file_name
        else:
            out_file = Path(f"{out_file}.state")

        with open(out_file, "w") as f:
            f.writelines(state_file_lines)
